@@ -1,4 +1,4 @@
-__version__ = '000.000.047'
+__version__ = '000.000.048'
 __author__ = 'Calafiore Carmelo'
 __author_email__ = 'c.calafiore@essex.ac.uk'
 __maintainer_email__ = 'c.calafiore@essex.ac.uk'
@@ -7,8 +7,6 @@ __all__ = [
     'format_args', 'image', 'maths', 'mixamo', 'nn', 'plot', 'preprocessing', 'shutdown', 'stats',
     'stimulation', 'strings', 'txt']
 
-# from . import *
-
 
 def initiate(names_submodules=None):
     template_import_string = "from . import "
@@ -23,11 +21,12 @@ def initiate(names_submodules=None):
                 import_string_m = template_import_string + str(name_m)
             exec(import_string_m)
     elif names_submodules is None:
-        for name_m in __all__:
-            if isinstance(name_m, str):
-                import_string_m = template_import_string + name_m
-            else:
-                import_string_m = template_import_string + str(name_m)
-            exec(import_string_m)
+        exec('from . import *')
+        # for name_m in __all__:
+        #     if isinstance(name_m, str):
+        #         import_string_m = template_import_string + name_m
+        #     else:
+        #         import_string_m = template_import_string + str(name_m)
+        #     exec(import_string_m)
     else:
         raise TypeError(names_submodules)
