@@ -14,6 +14,7 @@ import os
 import copy
 import datetime
 from . import array as cc_array
+from . import clock as cc_clock
 from . import maths as cc_maths
 from . import combinations as cc_combinations
 from .strings import format_float_to_str as cc_strings_format_float_to_str
@@ -127,13 +128,14 @@ class Loader:
             self.batch_size = batch_size
             self.n_batches = cc_maths.rint(self.n_samples / self.batch_size)
 
-        # print('self.batch_size =', self.batch_size)
-        # print('self.n_samples =', self.n_samples)
-        # print('self.n_batches =', self.n_batches)
-        # print('self.time.range =', self.time.range)
-        # print('self.n_conditions_directories =', self.n_conditions_directories)
+        print('self.batch_size =', self.batch_size)
+        print('self.n_samples =', self.n_samples)
+        print('self.n_batches =', self.n_batches)
+        print('self.time.range =', self.time.range)
+        print('self.n_conditions_directories =', self.n_conditions_directories)
         # print('self.conditions_directories =', self.conditions_directories)
 
+
         if self.shuffle:
             self.batches_indexes = None
         else:
@@ -167,17 +169,21 @@ class Loader:
         self.shape_sample = shape_sample_0
         self.shape_batch = torch.Size(tuple([self.batch_size] + shape_sample_0))
         self.tensor_batch = torch.empty(self.shape_batch, dtype=tensor_0.dtype, device=device)
+
         self.n_dims_samples = self.shape_sample.__len__()
         self.n_dims_batch = self.n_dims_samples + 1
         self.indexes_batch = np.empty(self.n_dims_batch, dtype=object)
         self.indexes_batch[1:] = slice(0, None, 1)
 
+        print('self.tensor_batch.shape =', self.tensor_batch.shape)
+        print()
+
     def load_batches_e(self):
 
         if self.shuffle:
             self.batches_indexes = np.split(np.random.permutation(self.n_samples), self.n_batches, axis=0)
-        # for b in range(3):
-        for b in range(self.n_batches):
+        for b in range(3):
+        # for b in range(self.n_batches):
             labels_b = self.labels[self.batches_indexes[b]]
             # batch_size_eb = batches_indexes_e.shape[0]
             for i in range(self.batch_size):
@@ -209,7 +215,7 @@ class Loader:
 def train_early_stop(model, loader, criterion, optimizer, scheduler, I=10):
 
     # timer = cc_clock.Timer()
-    datetime_start = datetime.datetime.today()
+    cc_timer = cc_clock.AllUnitsTimer()
 
     for key_loader_k in loader.keys():
         if key_loader_k == 'training' or key_loader_k == 'validation':
@@ -219,7 +225,9 @@ def train_early_stop(model, loader, criterion, optimizer, scheduler, I=10):
 
     headers = [
         'Epoch', 'Unsuccessful Epochs', 'Training Loss', 'Training Accuracy',
-        'Valuation Loss', 'Valuation Accuracy', 'Best Valuation Accuracy', 'Is Better Accuracy']
+        'Valuation Loss', 'Lowest Valuation Loss', 'Is Lower Loss',
+        'Valuation Accuracy', 'Highest Valuation Accuracy', 'Is Higher Accuracy']
+
     n_columns = len(headers)
     new_line_stats = [None] * n_columns  # type: list
 
@@ -231,15 +239,22 @@ def train_early_stop(model, loader, criterion, optimizer, scheduler, I=10):
     n_decimals_for_printing = 6
 
     best_model_wts = copy.deepcopy(model.state_dict())
-    best_acc = 0.0
-    best_acc_str = cc_strings_format_float_to_str(best_acc, n_decimals=n_decimals_for_printing)
+
+    lowest_loss = math.inf
+    lowest_loss_str = str(lowest_loss)
+    highest_accuracy = -math.inf
+    highest_accuracy_str = str(highest_accuracy)
 
     i = 0
     e = 0
 
+    n_dashes = 110
+    dashes = '-' * n_dashes
+    print(dashes)
+
     while i < I:
 
-        print('Epoch {} - Unsuccessful Epochs {}'.format(e, i) + '\n' + ('-' * 10))
+        print('Epoch {} - Unsuccessful Epochs {}'.format(e, i))
 
         stats['lines'].append(new_line_stats.copy())
         stats['lines'][e][stats['headers']['Epoch']] = e
@@ -276,10 +291,10 @@ def train_early_stop(model, loader, criterion, optimizer, scheduler, I=10):
             # loss_float_eb = loss_eb.item()
             # # noinspection PyTypeChecker
             # corrects_eb = torch.sum(preds == labels_eb).item()
-            # acc_eb = corrects_eb / batch_eb.shape[0]
+            # accuracy_eb = corrects_eb / batch_eb.shape[0]
             # loss_str_eb = cc_strings_format_float_to_str(loss_float_eb, n_decimals=n_decimals_for_printing)
-            # acc_str_eb = cc_strings_format_float_to_str(acc_eb, n_decimals=n_decimals_for_printing)
-            # print('Training. Epoch: {:d}. Batch {:d}. Loss: {:s}. Acc: {:s}.'.format(e, b, loss_str_eb, acc_str_eb))
+            # accuracy_str_eb = cc_strings_format_float_to_str(accuracy_eb, n_decimals=n_decimals_for_printing)
+            # print('Training. Epoch: {:d}. Batch {:d}. Loss: {:s}. Accuracy: {:s}.'.format(e, b, loss_str_eb, accuracy_str_eb))
 
             running_loss_e += loss_eb.item() * batch_eb.shape[0]
             # noinspection PyTypeChecker
@@ -290,20 +305,19 @@ def train_early_stop(model, loader, criterion, optimizer, scheduler, I=10):
         # scheduler.step()
 
         loss_e = running_loss_e / loader['training'].n_samples
-        acc_e = running_corrects_e / loader['training'].n_samples
+        accuracy_e = running_corrects_e / loader['training'].n_samples
         # loss_e = float(running_loss_e / loader['training'].n_samples)
-        # acc_e = float(running_corrects_e / loader['training'].n_samples)
+        # accuracy_e = float(running_corrects_e / loader['training'].n_samples)
 
         loss_str_e = cc_strings_format_float_to_str(loss_e, n_decimals=n_decimals_for_printing)
-        acc_str_e = cc_strings_format_float_to_str(acc_e, n_decimals=n_decimals_for_printing)
+        accuracy_str_e = cc_strings_format_float_to_str(accuracy_e, n_decimals=n_decimals_for_printing)
 
-        print('Training. Epoch: {:d}. Loss: {:s}. Acc: {:s}. Best Acc: {:s}.'.format(
-            e, loss_str_e, acc_str_e, best_acc_str))
+        print('Epoch: {:d}. Training.   Loss: {:s}. Accuracy: {:s}'.format(e, loss_str_e, accuracy_str_e))
 
         stats['lines'][e][stats['headers']['Training Loss']] = loss_e
-        stats['lines'][e][stats['headers']['Training Accuracy']] = acc_e
+        stats['lines'][e][stats['headers']['Training Accuracy']] = accuracy_e
         # stats['Training Loss'].append(float(loss_e))
-        # stats['Training Accuracy'].append(float(acc_e))
+        # stats['Training Accuracy'].append(float(accuracy_e))
 
 
         # validation phase
@@ -330,10 +344,10 @@ def train_early_stop(model, loader, criterion, optimizer, scheduler, I=10):
             # loss_float_eb = loss_eb.item()
             # # noinspection PyTypeChecker
             # corrects_eb = torch.sum(preds == labels_eb).item()
-            # acc_eb = corrects_eb / batch_eb.shape[0]
+            # accuracy_eb = corrects_eb / batch_eb.shape[0]
             # loss_str_eb = cc_strings_format_float_to_str(loss_float_eb, n_decimals=n_decimals_for_printing)
-            # acc_str_eb = cc_strings_format_float_to_str(acc_eb, n_decimals=n_decimals_for_printing)
-            # print('Validation. Epoch: {:d}. Batch {:d}. Loss: {:s}. Acc: {:s}.'.format(e, b, loss_str_eb, acc_str_eb))
+            # accuracy_str_eb = cc_strings_format_float_to_str(accuracy_eb, n_decimals=n_decimals_for_printing)
+            # print('Validation. Epoch: {:d}. Batch {:d}. Loss: {:s}. Accuracy: {:s}.'.format(e, b, loss_str_eb, accuracy_str_eb))
 
             running_loss_e += loss_eb.item() * batch_eb.shape[0]
             # noinspection PyTypeChecker
@@ -342,216 +356,70 @@ def train_early_stop(model, loader, criterion, optimizer, scheduler, I=10):
             b += 1
 
         loss_e = running_loss_e / loader['validation'].n_samples
-        acc_e = running_corrects_e / loader['validation'].n_samples
+        accuracy_e = running_corrects_e / loader['validation'].n_samples
         # loss_e = float(running_loss_e / loader['training'].n_samples)
-        # acc_e = float(running_corrects_e / loader['training'].n_samples)
+        # accuracy_e = float(running_corrects_e / loader['training'].n_samples)
 
         loss_str_e = cc_strings_format_float_to_str(loss_e, n_decimals=n_decimals_for_printing)
-        acc_str_e = cc_strings_format_float_to_str(acc_e, n_decimals=n_decimals_for_printing)
+        accuracy_str_e = cc_strings_format_float_to_str(accuracy_e, n_decimals=n_decimals_for_printing)
 
-        print('Validation. Epoch: {:d}. Loss: {:s}. Acc: {:s}. Best Acc: {:s}.'.format(
-            e, loss_str_e, acc_str_e, best_acc_str))
+        print('Epoch: {:d}. Validation. Loss: {:s}. Lowest Loss {:s}. Accuracy: {:s}. Highest Accuracy: {:s}.'.format(
+            e, loss_str_e, lowest_loss_str, accuracy_str_e, highest_accuracy_str))
 
         stats['lines'][e][stats['headers']['Valuation Loss']] = loss_e
-        stats['lines'][e][stats['headers']['Valuation Accuracy']] = acc_e
-        stats['lines'][e][stats['headers']['Best Valuation Accuracy']] = best_acc
-        # stats['Valuation Loss'].append(float(loss_e))
-        # stats['Valuation Accuracy'].append(float(acc_e))
-        # stats['Best Valuation Accuracy'].append(best_acc)
+        stats['lines'][e][stats['headers']['Lowest Valuation Loss']] = lowest_loss
+        stats['lines'][e][stats['headers']['Valuation Accuracy']] = accuracy_e
+        stats['lines'][e][stats['headers']['Highest Valuation Accuracy']] = highest_accuracy
 
-        if acc_e > best_acc:
-            stats['lines'][e][stats['headers']['Is Better Accuracy']] = 1
-            # stats['Is Better Accuracy'].append(1)
+        if loss_e < lowest_loss:
+            stats['lines'][e][stats['headers']['Is Lower Loss']] = 1
             i = 0
-            best_acc = acc_e
-            best_acc_str = cc_strings_format_float_to_str(best_acc, n_decimals=n_decimals_for_printing)
+            lowest_loss = loss_e
+            lowest_loss_str = cc_strings_format_float_to_str(lowest_loss, n_decimals=n_decimals_for_printing)
             best_model_wts = copy.deepcopy(model.state_dict())  # deep copy the model
         else:
-            stats['lines'][e][stats['headers']['Is Better Accuracy']] = 0
-            # stats['Is Better Accuracy'].append(0)
+            stats['lines'][e][stats['headers']['Is Lower Loss']] = 0
             i += 1
 
-        e += 1
-        print()
+        if accuracy_e > highest_accuracy:
+            stats['lines'][e][stats['headers']['Is Higher Accuracy']] = 1
+            highest_accuracy = accuracy_e
+            highest_accuracy_str = cc_strings_format_float_to_str(highest_accuracy, n_decimals=n_decimals_for_printing)
+        else:
+            stats['lines'][e][stats['headers']['Is Higher Accuracy']] = 0
 
+        e += 1
+        print(dashes)
+
+    print()
     E = e
 
-    datetime_end = datetime.datetime.today()
-    time_elapsed = datetime_end - datetime_start
+    time_training = cc_timer.get_all_unit_time()
 
-    # all_seconds = timer.get_time()
-    all_seconds = time_elapsed.seconds
-    seconds = all_seconds % 60
-    all_minutes = math.floor(all_seconds / 60)
-    minutes = all_minutes % 60
-    hours = math.floor(all_minutes / 60)
-
-    print('Training complete in {d:d}d {h:d}h {m:d}m {s:d}s'.format(
-        d=time_elapsed.days, h=hours, m=minutes, s=seconds))
-    print('Number of Epochs is {E:d}'.format(E=E))
-    print('Best Validation Acc: {:s}'.format(best_acc_str))
+    print('Training complete in {d:s} days {h:s} hours {m:s} minutes {s:s} seconds'.format(
+        d=str(time_training.days), h=str(time_training.hours),
+        m=str(time_training.minutes), s=str(time_training.seconds)))
+    print('Number of Epochs: {E:d}'.format(E=E))
+    print('Lowest Validation Loss: {:s}'.format(lowest_loss_str))
+    print('Highest Validation Accuracy: {:s}'.format(highest_accuracy_str))
 
     # load best model weights
     model.load_state_dict(best_model_wts)
     return model, stats
 
 
-def train_early_stop_old(model, loader, criterion, optimizer, scheduler, I=10):
-
-    # timer = cc_clock.Timer()
-    datetime_start = datetime.datetime.today()
-
-    best_model_wts = copy.deepcopy(model.state_dict())
-    best_acc = 0.0
-
-    headers = [
-        'Epoch', 'Unsuccessful Epochs', 'Training Loss', 'Training Accuracy',
-        'Valuation Loss', 'Valuation Accuracy', 'Best Valuation Accuracy', 'Is Better Accuracy']
-    n_columns = len(headers)
-    new_line_stats = [None] * n_columns  # type: list
-
-    stats = {
-        'headers': {headers[k]: k for k in range(n_columns)},
-        'n_columns': n_columns,
-        'lines': []}
-
-    i = 0
-    e = 0
-
-    while i < I:
-
-        print('Epoch {} - Unsuccessful Epochs {}'.format(e, i) + '\n' + ('-' * 10))
-
-        stats['lines'].append(new_line_stats.copy())
-        stats['lines'][e][stats['headers']['Epoch']] = e
-        stats['lines'][e][stats['headers']['Unsuccessful Epochs']] = i
-
-        # Each epoch has a training and a validation phase
-        for phase in loader.keys():
-            if phase == 'training':
-                model.train()  # Set model to training mode
-            elif phase == 'validation':
-                model.eval()  # Set model to validation mode
-            else:
-                raise ValueError('Unknown phase')
-
-            running_loss_e = 0.0
-            running_corrects_e = 0.0  # type: torch.tensor
-
-            b = 0
-            # Iterate over data.
-            for batch_eb, labels_eb in loader[phase].load_batches_e():
-
-                # zero the parameter gradients
-                optimizer.zero_grad()
-
-                # forward
-                # track history if only in training
-                with torch.set_grad_enabled(phase == 'training'):
-                    outputs = model(batch_eb)
-                    _, preds = torch.max(outputs, 1)
-                    loss = criterion(outputs, labels_eb)
-
-                    # backward + optimize only if in training phase
-                    if phase == 'training':
-                        loss.backward()
-                        optimizer.step()
-
-                # statistics
-                running_loss_eb = loss.item() * batch_eb.size(0)
-                # noinspection PyTypeChecker
-                running_corrects_eb = torch.sum(preds == labels_eb)
-
-                loss_eb = running_loss_eb / loader[phase].batch_size
-                acc_eb = running_corrects_eb.double() / loader[phase].batch_size
-                # epoch_acc = running_corrects.double() / loader[phase].n_samples
-
-                print('{}. Epoch: {:d}. Batch {:d}. Loss: {:s}. Acc: {:s}.'.format(phase, e, b, loss_eb, acc_eb))
-
-                running_loss_e += running_loss_eb
-                running_corrects_e += running_corrects_eb
-
-                b += 1
-
-            # if phase == 'training':
-            #     scheduler.step()
-
-            loss_e = float(running_loss_e / loader[phase].n_samples)
-            acc_e = float(running_corrects_e / loader[phase].n_samples)
-            # epoch_acc = running_corrects.double() / loader[phase].n_samples
-
-            print('\n{}. Epoch: {:d}. Loss: {:s}. Acc: {:s}. Best Acc: {:s}.'.format(
-                phase, e, loss_e, acc_e, best_acc))
-
-            if phase == 'training':
-                stats['lines'][e][stats['headers']['Training Loss']] = loss_e
-                stats['lines'][e][stats['headers']['Training Accuracy']] = acc_e
-                # stats['Training Loss'].append(float(loss_e))
-                # stats['Training Accuracy'].append(float(acc_e))
-            elif phase == 'validation':
-                stats['lines'][e][stats['headers']['Valuation Loss']] = loss_e
-                stats['lines'][e][stats['headers']['Valuation Accuracy']] = acc_e
-                stats['lines'][e][stats['headers']['Best Valuation Accuracy']] = best_acc
-
-                # stats['Valuation Loss'].append(float(loss_e))
-                # stats['Valuation Accuracy'].append(float(acc_e))
-                # stats['Best Valuation Accuracy'].append(best_acc)
-
-                if acc_e > best_acc:
-                    stats['lines'][e][stats['headers']['Is Better Accuracy']] = 1
-
-                    # stats['Is Better Accuracy'].append(1)
-                    i = 0
-                    best_acc = acc_e
-                    best_model_wts = copy.deepcopy(model.state_dict())  # deep copy the model
-                else:
-                    stats['lines'][e][stats['headers']['Is Better Accuracy']] = 0
-
-                    # stats['Is Better Accuracy'].append(0)
-                    i += 1
-            else:
-                raise ValueError('Unknown phase')
-
-        e += 1
-        print()
-
-    E = e
-
-    datetime_end = datetime.datetime.today()
-    time_elapsed = datetime_end - datetime_start
-
-    # all_seconds = timer.get_time()
-    all_seconds = time_elapsed.seconds
-    seconds = all_seconds % 60
-    all_minutes = math.floor(all_seconds / 60)
-    minutes = all_minutes % 60
-    hours = math.floor(all_minutes / 60)
-
-    print('Training complete in {d:d}d {h:d}h {m:d}m {s:d}s'.format(
-        d=time_elapsed.days, h=hours, m=minutes, s=seconds))
-    print('Number of Epochs is {E:d}'.format(E=E))
-    print('Best Validation Acc: {:s}'.format(best_acc))
-
-    # load best model weights
-    model.load_state_dict(best_model_wts)
-    return model, stats
-
-
-def load_resnet(name_model, K=None, pretrained=False, device=None):
+def load_model(name_model, K=None, pretrained=False, device=None):
     # model = torch.hub.load('pytorch/vision:v0.6.0', 'resnet152', pretrained=True)
 
-    if name_model == 'resnet18':
-        model = torchvision.models.resnet18(pretrained=pretrained)
-    elif name_model == 'resnet34':
-        model = torchvision.models.resnet34(pretrained=pretrained)
-    elif name_model == 'resnet50':
-        model = torchvision.models.resnet50(pretrained=pretrained)
-    elif name_model == 'resnet101':
-        model = torchvision.models.resnet101(pretrained=pretrained)
-    elif name_model == 'resnet152':
-        model = torchvision.models.resnet152(pretrained=pretrained)
+    if isinstance(name_model, str):
+        template_load_string = f"model = torchvision.models.{name_model:s}(pretrained=pretrained)"
     else:
-        raise ValueError('name_model')
+        template_load_string = f"model = torchvision.models.{str(name_model):s}(pretrained=pretrained)"
+
+    dict_globals = {'__builtins__': None}
+    dict_locals = {'torchvision': torchvision, 'pretrained': pretrained}
+    exec(template_load_string, dict_globals, dict_locals)
+    model = dict_locals['model']
 
     if K is not None:
         num_ftrs = model.fc.in_features
